"""A module to format data as a csv file."""

import csv
import io
from typing import Any, Dict, List


def format_csv(data: List[Dict[str, Any]], with_header: bool = False) -> str:
    """Format a list of dictionaries as a csv file.

    :param data: A list of dictionaries to format as a csv file.
    :param with_header: Whether to include the header in the csv file.

    :return: A string formatted as a csv file.
    """
    if len(data) == 0:
        return ""

    csv_data = io.StringIO()
    writer = csv.DictWriter(csv_data, fieldnames=data[0].keys())
    if with_header:
        writer.writeheader()
    writer.writerows(data)

    return csv_data.getvalue()



"""Geneset CLI commands."""

# ruff: noqa: B008
import json
from typing import List, Optional

import typer
from geneweaver.client.api import aon, genesets, mapping
from geneweaver.client.auth import get_access_token
from geneweaver.client.utils.cli.print.csv import format_csv
from geneweaver.core.enum import GeneIdentifier

cli = typer.Typer()

HELP_MESSAGE = """
The geneset commands allow you to authenticate with the GeneWeaver API.
"""


@cli.command()
def get(
    ctx: typer.Context,
    geneset_id: int,
    gene_id_type: Optional[GeneIdentifier] = typer.Option(None, case_sensitive=False),
    as_csv: bool = typer.Option(False, "--csv", help="Output as CSV"),
) -> dict:
    """Get a Geneset by ID."""
    result = genesets.get(get_access_token(), geneset_id, gene_id_type=gene_id_type)
    if not ctx.obj["quiet"]:
        if ctx.obj["pretty"]:
            typer.echo(json.dumps(result, indent=4))
        else:
            typer.echo(json.dumps(result))
    return result


@cli.command()
def get_values_as_ensembl_mouse(
    ctx: typer.Context,
    geneset_id: int,
    in_threshold: Optional[bool] = None,
    algorithm: Optional[aon.OrthologAlgorithms] = typer.Option(
        default=None, help="Ortholog mapping algorithm. Leave empty for all algorithms."
    ),
    as_csv: bool = typer.Option(False, "--csv", help="Output as CSV"),
) -> List[dict]:
    """Get a Geneset's values as Ensembl Mouse Gene IDs."""
    # Check Geneset Species
    token = get_access_token()
    result = mapping.ensembl_mouse_mapping(
        token,
        geneset_id,
        in_threshold,
        algorithm,
    )

    if as_csv:
        result = format_csv(result)

    if not ctx.obj["quiet"]:
        if not as_csv:
            if ctx.obj["pretty"]:
                typer.echo(json.dumps(result, indent=4))
            else:
                typer.echo(json.dumps(result))
        else:
            typer.echo(result)

    return result



"""Utilities for AON interaction."""

from collections.abc import Hashable
from typing import Dict, List, Tuple, TypeVar, Union

Number = TypeVar("Number", bound=Union[int, float])
T = TypeVar("T", bound=Hashable)


def map_symbols(
    original: Dict[T, Number], mappings: List[Tuple[T, T]]
) -> Dict[T, Number]:
    """Map gene symbols using a list of mappings.

    When a human gene maps to multiple mouse genes, there should be a row for each
    of those mouse ids to be included, and therefore the score will be the same
    for all of those.

    When multiple human genes map to the same mouse gene, we want to keep the
    gene that has the highest abs(score).

    :param original: The original gene symbols and values.
                     With the key as the identifier and the value as the value.
    :param mappings: The mappings to use.
                     With the source as the first item and the target as the second.

    :return: The mapped gene symbols and values.
    """
    mapped_values = {}
    for mapping in mappings:
        source, target = mapping

        try:
            value = original[source]
        except KeyError:
            continue

        if target not in mapped_values or abs(value) > abs(mapped_values[target]):
            mapped_values[target] = value

    return mapped_values



"""Functions that warp the Geneweaver /genes endpoint."""

from typing import List

from geneweaver.client.api.utils import format_endpoint, sessionmanager
from geneweaver.core.enum import GeneIdentifier, Species

ENDPOINT = "genes"


def mappings(
    access_token: str,
    source_ids: List[str],
    target_id_type: GeneIdentifier,
    species: Species,
) -> dict:
    """Get mappings for a list of genes.

    :param access_token: User access token
    :param source_ids: List of source gene IDs.
    :param target_id_type: Target gene ID type (one of GeneIdentifier).
    :param species: Species of the identifiers.
    """
    with sessionmanager(token=access_token) as session:
        resp = session.post(
            format_endpoint(ENDPOINT, "mappings"),
            json={
                "source_ids": source_ids,
                "target_gene_id_type": str(target_id_type),
                "species": str(species),
            },
        )

    return resp.json()



"""Functions that wrap the GeneWeaver API on /genesets endpoints."""

from enum import Enum
from typing import List, Optional

from geneweaver.client.api.utils import sessionmanager
from geneweaver.client.core.config import settings
from geneweaver.core.enum import Species


class OrthologAlgorithms(Enum):
    """The available ortholog algorithms in Geneweaver AON."""

    HGNC = "HGNC"
    PANTHER = "PANTHER"
    HIERANOID = "Hieranoid"
    PHYLOMEDB = "PhylomeDB"
    ORTHOINSPECTOR = "OrthoInspector"
    INPARANOID = "InParanoid"
    ORTHOFINDER = "OrthoFinder"
    ZFIN = "ZFIN"
    ENSEMBLCOMPARA = "EnsemblCompara"
    SONICPARANOID = "SonicParanoid"
    OMA = "OMA"
    XENBASE = "Xenbase"


def ortholog_mapping(
    identifiers: List[str], to_species: Species, algorithm_id: Optional[int] = None
) -> dict:
    """Get ortholog mapping for a list of genes.

    :param identifiers: List of gene values (must be AON ID type for Species).
    :param to_species: Species to map to.
    :param algorithm_id: Algorithm ID for mapping.

    :return: Ortholog mapping dict.
    """
    params = {"to_species": int(to_species), "limit": 30000}

    if algorithm_id is not None:
        params["algorithm_id"] = algorithm_id

    with sessionmanager() as session:
        resp = session.post(
            settings.AON_API_URL + "/genes/ortholog/mapping",
            params=params,
            json=identifiers,
        )
    return resp.json()


def algorithm_id_from_name(algorithm_name: str) -> int:
    """Get algorithm ID from algorithm name.

    :param algorithm_name: The name of the algorithm.
    :return: The algorithm ID.
    """
    with sessionmanager() as session:
        resp = session.get(settings.AON_API_URL + "/algorithms")
        algorithms = resp.json()
        for algorithm in algorithms:
            if algorithm["alg_name"].lower().replace(
                " ", ""
            ) == algorithm_name.lower().replace(" ", ""):
                return algorithm["alg_id"]



"""Geneweaver API CLI for Geneset endpoints."""

# ruff: noqa: B008
import json
from typing import Optional

import typer
from geneweaver.client.api import genesets
from geneweaver.client.auth import get_access_token
from geneweaver.core.enum import GeneIdentifier

cli = typer.Typer()

HELP_MESSAGE = """
Tools and utilities for interacting with the Geneweaver API.
"""


@cli.command()
def get(
    ctx: typer.Context,
    geneset_id: int,
    gene_id_type: Optional[GeneIdentifier] = typer.Option(None, case_sensitive=False),
) -> dict:
    """Get a Geneset by ID."""
    result = genesets.get(get_access_token(), geneset_id, gene_id_type=gene_id_type)
    if not ctx.obj["quiet"]:
        if ctx.obj["pretty"]:
            typer.echo(json.dumps(result, indent=4))
        else:
            typer.echo(json.dumps(result))
    return result



"""A module to interact with the graph service.

This module contains a client API for interacting with the
graph service. The find and search operations on the graph
are supported. The find is used for variant mapping and can
include long-running searches, for instance 100k variants
mapped from one species to another. The search contains simpler
and shorter running operations, for instance to get orthologs.
This intentionally obfuscates the underlying graph database
for the reasons of security and scalability.
"""

# ruff: noqa: E501, N803
import csv
import json
from enum import Enum
from typing import Optional

import requests

Source = Enum("Source", ["BAYLOR", "HOMOLOGENE" "ENSEMBL"])
"""
@usage: from geneweaver.client.api.graph import Source
"""

SearchType = Enum(
    "SearchType",
    [
        "EQTL_BP_TO_GENES",
        "EQTL_TISSUE_TO_GENES",
        "GENE_ID",
        "PEAKS",
        "EQTL_BP_TO_PEAKS",
        "ORTHOLOG",
    ],
)
"""
@usage: from geneweaver.client.api.graph import SearchType
"""


class SearchRequest:
    """A class to hold the search request properties.

    This is just a type to give a prospective search
    subroutine user a fighting chance at knowing the
    available properties.
    """

    def __init__(
        self,
        searchType: SearchType = SearchType.EQTL_BP_TO_GENES,
        delimiter: str = ",",
        species: str = "Mus musculus",
        geneIds: Optional[list] = None,
        eqtls: Optional[list] = None,
        rsId: Optional[str] = None,
        tissue: Optional[str] = None,
        studyId: Optional[str] = None,
        source: Optional[Source] = None,
    ) -> None:
        """Initialize a SearchRequest object."""
        # Defaulted properties
        self.searchType: SearchType = searchType
        self.delimiter: str = delimiter
        self.species: str = species

        # Available properties
        self.geneIds: list = geneIds
        self.eqtls: list = eqtls
        self.rsId: str = rsId
        self.tissue: str = tissue
        self.studyId: str = studyId
        self.source: Source = source  # BAYLOR, ENSEMBL, HOMOLOGENE


class Position:
    """A class to hold a position on a chromosome."""

    def __init__(self, bp: str, chr: str) -> None:  # noqa: A002
        """Initialize a Position object."""
        self.bp = bp
        self.chr = chr


class VariantGraphClient:
    """A client to wrap calling the variant graph API."""

    def __init__(self, url: str = "https://graph-api.geneweaver.org/") -> None:
        """Create a VariantGraphClient from a URL.

        :param url: The URL to which we will connect when making graph server queries.
        TODO Authentication?
        """
        self.url = url

    def heartbeat(self) -> dict:
        """Check the server connection.

        Check if we are connected to a server which is in turn connected to a graph
        database.

        :return: a dict with parameters to define if connected and to where.
        """
        response = requests.get(self._get_connected_url())
        if not response.ok:
            response.raise_for_status()
        return json.loads(response.text)

    def get_orthologs(
        self, geneset: list, species: str = "Mus musculus", source: Source = None
    ) -> dict:
        """Get the orthologs for a geneset.

        The species is the species to which we are mapping. So if a geneid in the
        geneset is human it will find the mouse mapped gene. If the geneid is
        already mouse then no mapping will be found.

        :param geneset: list of geneIds
        :param species: String for species to which we map, default of mouse.
                        Must be binomial name, case-sensitive.
        :param source: None, BAYLOR, HOMOLOGENE or ENSEMBL.
                        To get homologs use source = Source.HOMOLOGENE

        :return: all the orthologs found in the geneset which map.
                If there are geneids of the target species, these are not in the dict.
        """
        request = {
            "geneIds": geneset,
            "searchType": "ORTHOLOG",  # Even homologs are 'searchType' ORTHOLOG
            "species": species,
        }  # species to map to

        if source is not None:
            request["source"] = str(source)

        csv_lines = self.search(request)

        reader = csv.reader(csv_lines)
        omap = {
            row[0]: row[1]
            for row in reader
            if len(row) > 0 and not row[0].startswith("gf.geneId")
        }

        return omap

    def search(self, request: SearchRequest) -> list:
        """Perform a search.

        :param request: object with fields defining possible search.
        :return: list of csv values depending on the searchtype.

        Search using a SearchRequest to return as csv. In the returned
        results 'g' is the code for the gene, 'e' the eqtl, 'v' the variant and 'p'Â the peak.

        So for instance a result of:
            g.chr    g.geneName    g.geneId    e.tissueName    e.tissueGroup    e.bp    e.uberon    e.lod    v.start    v.end    v.rsId
            3    Rpsa-ps10    ENSMUSG00000047676    bone    null    101555002    null    6.069993732615    101204207    101204207    rs31201108
            3    Rpsa-ps10    ENSMUSG00000047676    bone    null    101555002    null    6.069993732615    101204207    101204207    rs31201108
            3    Rpsa-ps10    ENSMUSG00000047676    bone    null    101555002    null    6.069993732615    101204207    101204207    rs31201108
            For EQTL the 'e.bp' will in bp38/mm10 and the other coordinates are in bp39/mm11. In general the coordinates of
            the underlying database and the results are bp39/mm11."

            - Examples
            - 1. Given an eQTL in b38 coordinates or rsid, find gene names, gene ids mapped to this eQTL, tissue, UBERON_id, and lod score.
            {eqtls: [{
                 bp : 101555002,
                 chr : 7
            }]}

            You can also provide the rsId if you know it:
            {eqts: [{
                 bp : 101555002,
                 chr : 7,
                }],"
                rsId: rs31201108
            }
            Other fields are available, such as species, see the 'SearchRequest' documentation below
            Even if the EQTL did not link to a variant you can find it:
            {eqtls: [{
                 bp : 9239543,
                 chr : 17
            }]}

            - 2. Given a gene name or gene id, find all eQTL variants associated to this gene.
            {
                 geneIds : [ENSMUSG00000037661],
                 searchType : GENE_ID
            }

            This also works for multiple gene ids in a single query:
            {
                 geneIds : [ENSMUSG00000037661, ENSMUSG00000044117, ENSMUSG00000065968],
                 searchType : GENE_ID
            }

            - 3. Given a peak type and peak location, find all variants in the peak.

        Todo: Fill out more details here
        ----
            - 4. Given an eQTL in b38 coordinates or rsid, find all peaks overlapping this variant.

        Todo: Fill out more details here
        ----

        """
        response = requests.post(self._get_search_url(), None, request)
        if not response.ok:
            response.raise_for_status()

        csv_lines = response.text.split("\n")
        return csv_lines

    def _get_search_url(self) -> str:
        return self.url + "variant/graph/search"

    def _get_connected_url(self) -> str:
        return self.url + "variant/graph/heartbeat"



"""Cross-API Geneset Symbol Mapping."""

from typing import List, Optional

from geneweaver.client.api import aon, genes, genesets
from geneweaver.client.utils.aon import map_symbols
from geneweaver.core.enum import GeneIdentifier, Species


def ensembl_mouse_mapping(
    access_token: str,
    geneset_id: int,
    in_threshold: bool,
    algorithm: Optional[aon.OrthologAlgorithms] = None,
) -> List[dict]:
    """Get a Geneset's values as Ensembl Mouse Gene IDs.

    :param access_token: User access token.
    :param geneset_id: Geneset ID.
    :param in_threshold: Whether to filter genes by threshold.
    :param algorithm: Ortholog mapping algorithm.

    :return: List of geneset values. `[{"symbol": k, "value": v}, ...]
    """
    response = genesets.get(access_token, geneset_id)
    species = Species(response["geneset"]["species_id"])

    gene_id_type = GeneIdentifier.ENSEMBLE_GENE

    if species == Species.HOMO_SAPIENS:
        gene_id_type = GeneIdentifier.HGNC

    response = genesets.get_values(access_token, geneset_id, gene_id_type, in_threshold)

    if species == Species.MUS_MUSCULUS:
        result = response["data"]

    else:
        if algorithm:
            algorithm_id = aon.algorithm_id_from_name(algorithm.value)
        else:
            algorithm_id = None

        aon_response = aon.ortholog_mapping(
            [g["symbol"] for g in response["data"]],
            Species.MUS_MUSCULUS,
            algorithm_id=algorithm_id,
        )

        mgi_result = map_symbols(
            {item["symbol"]: item["value"] for item in response["data"]},
            [(r["from_gene"], r["to_gene"]) for r in aon_response],
        )

        gw_map_response = genes.mappings(
            access_token,
            list(set(mgi_result.keys())),
            GeneIdentifier.ENSEMBLE_GENE,
            Species.MUS_MUSCULUS,
        )

        ensembl_result = map_symbols(
            mgi_result,
            [
                (r["original_ref_id"], r["mapped_ref_id"])
                for r in gw_map_response["gene_ids_map"]
            ],
        )

        result = [{"symbol": k, "value": v} for k, v in ensembl_result.items()]

    return result
